#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <nag.h>
#include <dco.hpp>
#include <nagad.h>
#include <stdio.h>
#include <math.h>
#include <nagx02.h>
#include <iostream>
using namespace std;


struct data {
    Integer m;
    Integer n;
    dco::ga1s<double>::type *X;
    dco::ga1s<double>::type *y;
};

struct data parameters;

template <typename T>
T f(const Integer &n, const T *p, const T *x){
    T res = 0.0;

    for (Integer i=0; i<n; ++i){
        res +=p[i]*x[i];
    }
    res = pow(res,2);
    return res;
}


template <typename T>
void E(const T *p, const struct data *par, T *fvec){
    Integer m = par->m;
    Integer n = par->n;
    T tmp;
    for (Integer i = 0; i < m; i++) {
        tmp = f(n, p, &par->X[i*n]);
        fvec[i] = tmp-par->y[i];
    }
}


template <typename T>
void lsqfun(void* &,
            Integer &,
            const Integer &,
            const Integer &,
            const T xc[],
            T fvec[],
            Integer *,
            T *)
{
    E(xc, &parameters, fvec);
    return;
}


template <typename T>
void lsqmon(void* &,
            const Integer &m,
            const Integer &n,
            const T xc[],
            const T fvec[],
            const T *,
            const Integer &,
            const T *,
            const Integer &,
            const Integer &niter,
            const Integer &,
            Integer *,
            T *)
{
    if (niter==0)
        printf("Iteration   Function Value          Current  Solution\n");

    double fsumsq = 0.0;
    for (int i=0;i<m; i++) {
        double f = dco::value(fvec[i]);
        fsumsq = fsumsq + f*f;
    }

    printf ("%5d       %10.5f", niter, fsumsq);
    size_t max_out = n;
    if (max_out>5) max_out=5;

    for (size_t i=0; i<max_out; i++)
        printf ("          %.5f", dco::value(xc[i]));

    printf("\n");

}


int main(int argc, char* argv[])
{
    using DCO_M=typename dco::ga1s<double>;
    using DCO_T=typename DCO_M::type;

    assert(argc==3); Integer m=std::stoi(argv[1]), n=std::stoi(argv[2]);

    Eigen::Matrix<double, Eigen::Dynamic, 1> y;
    Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> t;
    /* this is the data to be fitted */
    t = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(m,n);
    y = Eigen::Matrix<double, Eigen::Dynamic, 1>::Random(m);
    t=t.cwiseProduct(t);

    parameters.m = m; parameters.n = n;
    parameters.y = new DCO_T[m];
    for (Integer i=0; i<m; i++) parameters.y[i] = y(i);

    parameters.X = new DCO_T[m*n];
    for (Integer i=0; i<m; i++){
        for (Integer j=0; j<n; j++)
            parameters.X[i*n+j] = t(i,j);
    }

    // Scalars
    int               exit_status = 0;
    const Integer     ldfjac = m, ldv = n;
    Integer ifail = 0;

    // AD routine arguments
    Integer           iuser[1];
    Integer           iprint, maxcal, niter=0, nf=0;
    DCO_T eta, stepmx, xtol, fsumsq;
    DCO_T ruser[1];
    DCO_T *x = new DCO_T[n];
    DCO_T *fvec = new DCO_T[m];
    DCO_T *fjac = new DCO_T[m*n];
    DCO_T *s = new DCO_T[n];
    DCO_T *v = new DCO_T[ldv*n];

    // Create AD tape
    DCO_M::global_tape = DCO_M::tape_t::create();

    // Create AD configuration data object
    void    *ad_handle = 0;
    x10aa_a1w_f_(ad_handle,ifail);


    for (int i=0; i<m; i++) {
        DCO_M::global_tape->register_variable(parameters.y[i]);
    }

    iprint = 1;
    maxcal = 400*n;
    eta = 0.5;
    xtol = 10.0*sqrt(X02AJC);
    stepmx = 10.0;

    DCO_T *x_sav = new DCO_T[n];
    // Starting points
    for (Integer i=0; i<n; i++){
        x_sav[i] = -1.0;
        DCO_M::global_tape->register_variable(x_sav[i]);
        x[i] = x_sav[i];
    }

    ifail = -1;
    e04fc_a1w_f_(ad_handle,m,n,lsqfun,lsqmon,iprint,maxcal,eta,xtol,
                 stepmx,x,fsumsq,fvec,fjac,ldfjac,s,v,ldv,niter,nf,
                 iuser,ruser,ifail);

    cout << "\n  summary from method e04fc" << endl;
    cout << "  number of iterations: " << niter << endl;
    cout << "  function evaluations: " << nf << endl;
    // Primal results
    cout.setf(ios::fixed,ios::floatfield);
    if (ifail==0 || ifail>1) {
        cout.precision(6);
        cout << "\n  final |f(x)|     = ";
        cout.width(20); cout << dco::value(fsumsq) << endl;

        int max_out = n;
        if (max_out>5) max_out=5;
        cout << "  Minimum found at = ";
        for (int i=0; i<max_out; i++) {
            cout.width(20);
            cout << dco::value(x[i]);

        }
        cout << endl;

    }

    cout << "\n Derivatives calculated: First order adjoints\n";
    cout << " Derivatives:\n\n";

    // Setup evaluation of derivatives of fsumsq via adjoints.
    dco::derivative(fsumsq) += 1.0;
    DCO_M::global_tape->interpret_adjoint();

    // Get derivatives of fsumsq
    cout << "  sum of squares w.r.t y:\n\n";
    for (int i=0; i<m; i++) {
        double d = dco::derivative(parameters.y[i]);
        cout << "  dfsumsq/dy[" << i <<"]= " <<d << endl;
    }


    // Get derivatives of fsumsq
    cout << "\n  sum of squares w.r.t x:\n\n";
    for (int i=0; i<n; i++) {
        double d = dco::derivative(x_sav[i]);
        cout << "  dfsumsq/dx[" << i <<"]= " <<d << endl;
    }

    // Get derivatives of solution points w.r.t. y
    cout << "\n  solution points w.r.t y:\n";
    for (int j=0; j<n; j++) {
        DCO_M::global_tape->zero_adjoints();
        dco::derivative(x[j]) =1.0;
        DCO_M::global_tape->interpret_adjoint();
        cout << "\n dx(";
        cout.width(1); cout << j;
        cout << ")/dy :\n";
        for (int i=0; i<m; i++) {
            double d = dco::derivative(parameters.y[i]);
            cout.width(12); cout << d;
            if (i%5==4) {
                cout << endl;
            }
        }
    }
    cout << endl;
    // Remove computational data object and tape
    x10ab_a1w_f_(ad_handle,ifail);
    DCO_M::tape_t::remove(DCO_M::global_tape);

    delete [] parameters.X;
    delete [] parameters.y;
    delete [] x;
    delete [] x_sav;
    delete [] fvec;
    delete [] fjac;
    delete [] s;
    delete [] v;
    return exit_status;
}


